from time import perf_counter

import jax
from flax import linen as nn
from flax.serialization import to_bytes
from jax import numpy as jnp

from safejax import serialize


class SingleLayerModel(nn.Module):
    features: int

    @nn.compact
    def __call__(self, x):
        x = nn.Dense(features=self.features)(x)
        return x


model = SingleLayerModel(features=1)

rng = jax.random.PRNGKey(0)
params = model.init(rng, jnp.ones((1, 1)))


start_time = perf_counter()
for _ in range(100):
    serialize(params)
end_time = perf_counter()
print(f"safejax (100 runs): {end_time - start_time:0.4f} s")

start_time = perf_counter()
for _ in range(100):
    to_bytes(params)
end_time = perf_counter()
print(f"flax (100 runs): {end_time - start_time:0.4f} s")



from functools import partial
from pathlib import Path

from objax.variable import VarCollection
from safetensors import safe_open

from safejax.core.load import deserialize
from safejax.core.save import serialize  # noqa: F401
from safejax.typing import PathLike
from safejax.utils import OBJAX_VARIABLE_SEPARATOR

# `objax` params are usually defined as a `VarCollection`, and that's basically a dictionary with
# key-value pairs where the value is either a `BaseVar` or a `StateVar`. The problem is that when
# serializing those variables by default we just keep the value which is a `jnp.DeviceArray`, so we
# need to provide `include_objax_variables=True` to store the variable type names as part of the key
# using `::` as the separator. This is useful when deserializing the params, as we can restore a
# `VarCollection` object instead of a `Dict[str, jnp.DeviceArray]`.
serialize = partial(serialize, include_objax_variables=True)

# `objax` expects either a `Dict[str, jnp.DeviceArray]` or a `VarCollection` as model params
# which means any other type of `Dict` will not work. The only difference is that `VarCollection` can
# be assigned directly to `.vars()` while `Dict[str, jnp.DeviceArray]` needs to be manually assigned
# when looping over `.vars()`. Ideally, we want to restore the params from a `VarCollection`, that's why
# we've set the `to_var_collection` parameter to `True` by default.
deserialize = partial(deserialize, requires_unflattening=False, to_var_collection=True)


# When calling `deserialize` over an `objax.variable.VarCollection` object, those variables cannot
# be used directly for the inference, as the forward pass in `objax` is done through `__call__`, which
# implies that the class instance must contain the model params loaded in `.vars` attribute. So this
# function has been created in order to ease the parameter loading for `objax`, since as opposed to
# `flax` and `haiku`, the model params are not provided on every forward pass.
def deserialize_with_assignment(filename: PathLike, model_vars: VarCollection) -> None:
    """Deserialize a `VarCollection` from a file and assign it to a `VarCollection` object.

    Note:
        This function avoid some known issues related to the variable deserialization with `objax`,
        since the params are stored in a `VarCollection` object, which contains some `objax.variable`
        variables instead of key-tensor pais. So this way we avoid having to restore the `objax.variable`
        type per each value.

    Args:
        filename: Path to the file containing the serialized `VarCollection` as a `Dict[str, jnp.DeviceArray]`.
        model_vars: `VarCollection` object to which the deserialized tensors will be assigned.

    Returns:
        `None`, as the deserialized tensors are assigned to the `model_vars` object. So you
        just need to access `model_vars`, or the actual `model.vars()` attribute, since the
        assignment is done over a class attribute named `vars`.
    """
    if not isinstance(filename, (str, Path)):
        raise ValueError(
            "`filename` must be a `str` or a `pathlib.Path` object, not"
            f" {type(filename)}."
        )
    filename = filename if isinstance(filename, Path) else Path(filename)
    if not filename.exists or not filename.is_file:
        raise ValueError(f"`filename` must be a valid file path, not {filename}.")
    with safe_open(filename.as_posix(), framework="jax") as f:
        for key in f.keys():
            just_key = (
                key.split(OBJAX_VARIABLE_SEPARATOR)[0]
                if OBJAX_VARIABLE_SEPARATOR in key
                else key
            )
            if just_key not in model_vars.keys():
                raise ValueError(f"Variable with name {key} not found in model_vars.")
            model_vars[just_key].assign(f.get_tensor(key))



import jax
from flax import linen as nn
from flax.core.frozen_dict import FrozenDict
from jax import numpy as jnp

from safejax import deserialize, serialize


class SingleLayerModel(nn.Module):
    features: int

    @nn.compact
    def __call__(self, x):
        x = nn.Dense(features=self.features)(x)
        return x


network = SingleLayerModel(features=1)

rng_key = jax.random.PRNGKey(seed=0)
initial_params = network.init(rng_key, jnp.ones((1, 1)))

encoded_bytes = serialize(params=initial_params)
assert isinstance(encoded_bytes, bytes)
assert len(encoded_bytes) > 0

decoded_params = deserialize(path_or_buf=encoded_bytes, freeze_dict=True)
assert isinstance(decoded_params, FrozenDict)
assert len(decoded_params) > 0
assert decoded_params.keys() == initial_params.keys()

x = jnp.ones((1, 1))
y = network.apply(decoded_params, x)
assert y.shape == (1, 1)



import sys

import jax
from flax.serialization import to_bytes
from flaxmodels.resnet import ResNet50
from jax import numpy as jnp

from safejax import serialize

resnet50 = ResNet50()
params = resnet50.init(jax.random.PRNGKey(42), jnp.ones((1, 224, 224, 3)))


def serialization_safejax():
    _ = serialize(params)


def serialization_flax():
    _ = to_bytes(params)


if __name__ == "__main__":
    if len(sys.argv) < 2:
        raise ValueError("Please provide a function name to run as an argument")
    if sys.argv[1] not in globals():
        raise ValueError(f"Function {sys.argv[1]} not found")
    globals()[sys.argv[1]]()



"""`safejax `: Serialize JAX, Flax, Haiku, or Objax model params with `safetensors`"""

__author__ = "Alvaro Bartolome <alvarobartt@yahoo.com>"
__version__ = "0.5.0"

from safejax.core.load import deserialize  # noqa: F401
from safejax.core.save import serialize  # noqa: F401



import haiku as hk
import jax
from jax import numpy as jnp

from safejax import deserialize, serialize


def resnet_fn(x: jnp.DeviceArray, is_training: bool):
    resnet = hk.nets.ResNet50(num_classes=10)
    return resnet(x, is_training=is_training)


network = hk.without_apply_rng(hk.transform_with_state(resnet_fn))

rng_key = jax.random.PRNGKey(seed=0)
initial_params, initial_state = network.init(
    rng_key, jnp.ones([1, 224, 224, 3]), is_training=True
)

encoded_bytes = serialize(params=initial_params)
assert isinstance(encoded_bytes, bytes)
assert len(encoded_bytes) > 0

decoded_params = deserialize(path_or_buf=encoded_bytes)
assert isinstance(decoded_params, dict)
assert len(decoded_params) > 0
assert decoded_params.keys() == initial_params.keys()

x = jnp.ones([1, 224, 224, 3])
y, _ = network.apply(decoded_params, initial_state, x, is_training=False)
assert y.shape == (1, 10)



import warnings
from pathlib import Path
from typing import Union

from flax.core.frozen_dict import freeze
from fsspec import AbstractFileSystem
from objax.variable import VarCollection
from safetensors.flax import load, load_file

from safejax.typing import ParamsDictLike, PathLike
from safejax.utils import cast_objax_variables, unflatten_dict


def deserialize(
    path_or_buf: Union[PathLike, bytes],
    fs: Union[AbstractFileSystem, None] = None,
    freeze_dict: bool = False,
    requires_unflattening: bool = True,
    to_var_collection: bool = False,
) -> ParamsDictLike:
    """
    Deserialize JAX, Flax, Haiku, or Objax model params from either a `bytes` object or a file path,
    stored using `safetensors.flax.save_file` or directly saved using `safejax.save.serialize` with
    the `filename` parameter.

    Note:
        The default behavior of this function is to restore a `Dict[str, jnp.DeviceArray]` from
        a `bytes` object or a file path. If you are using `objax`, you should set `requires_unflattening`
        to `False` and `to_var_collection` to `True` to restore a `VarCollection`. If you're using `flax` you
        should set `freeze_dict` to `True` to restore a `FrozenDict`. Those are just tips on how to use it
        but all those frameworks are compatible with the default behavior.

    Args:
        path_or_buf:
            A `bytes` object or a file path containing the serialized model params.
        fs: The filesystem to use to load the model params. Defaults to `None`.
        freeze_dict:
            Whether to freeze the output `Dict` to be a `FrozenDict` or not. Defaults to `False`.
        requires_unflattening:
            Whether the model params require unflattening or not. Defaults to `True`.
        to_var_collection:
            Whether to convert the output `Dict` to a `VarCollection` or not. Defaults to `False`.

    Returns:
        A `Dict[str, jnp.DeviceArray]`, `FrozenDict`, or `VarCollection` containing the model params.
    """
    if isinstance(path_or_buf, bytes):
        decoded_params = load(data=path_or_buf)
    elif isinstance(path_or_buf, (str, Path)):
        if fs and fs.protocol != "file":
            if not isinstance(fs, AbstractFileSystem):
                raise ValueError(
                    "`fs` must be a `fsspec.AbstractFileSystem` object or `None`,"
                    f" not {type(fs)}."
                )
            with fs.open(path_or_buf, "rb") as f:
                decoded_params = load(data=f.read())
        else:
            if fs and fs.protocol == "file":
                filename = Path(fs._strip_protocol(path_or_buf))
            else:
                filename = (
                    path_or_buf if isinstance(path_or_buf, Path) else Path(path_or_buf)
                )
            if not filename.exists or not filename.is_file:
                raise ValueError(
                    f"`path_or_buf` must be a valid file path, not {path_or_buf}."
                )
            decoded_params = load_file(filename=filename.as_posix())
    else:
        raise ValueError(
            "`path_or_buf` must be a `bytes` object or a file path (`str` or"
            f" `pathlib.Path` object), not {type(path_or_buf)}."
        )
    if to_var_collection:
        try:
            return VarCollection(cast_objax_variables(params=decoded_params))
        except ValueError as e:
            warnings.warn(e)
        return decoded_params
    if requires_unflattening:
        decoded_params = unflatten_dict(params=decoded_params)
    if freeze_dict:
        return freeze(decoded_params)
    return decoded_params



from pathlib import Path
from typing import Dict, Union

from flax.core.frozen_dict import FrozenDict
from jax import numpy as jnp
from objax.variable import BaseVar, StateVar, VarCollection

PathLike = Union[str, Path]

JaxDeviceArrayDict = Dict[str, jnp.DeviceArray]
HaikuParams = Dict[str, JaxDeviceArrayDict]
ObjaxDict = Dict[str, Union[BaseVar, StateVar]]
ObjaxParams = Union[VarCollection, ObjaxDict]
FlaxParams = Union[Dict[str, Union[Dict, JaxDeviceArrayDict]], FrozenDict]

ParamsDictLike = Union[JaxDeviceArrayDict, HaikuParams, ObjaxParams, FlaxParams]



from jax import numpy as jnp
from objax.zoo.resnet_v2 import ResNet50

from safejax import deserialize, serialize

model = ResNet50(in_channels=3, num_classes=1000)

encoded_bytes = serialize(params=model.vars())
assert isinstance(encoded_bytes, bytes)
assert len(encoded_bytes) > 0

decoded_params = deserialize(
    encoded_bytes, requires_unflattening=False, to_var_collection=True
)
assert isinstance(decoded_params, dict)
assert len(decoded_params) > 0
assert decoded_params.keys() == model.vars().keys()

for key, value in decoded_params.items():
    if key not in model.vars():
        print(f"Key {key} not in model.vars()! Skipping.")
        continue
    model.vars()[key].assign(value)

x = jnp.ones((1, 3, 224, 224))
y = model(x, training=False)
assert y.shape == (1, 1000)



import os
import tempfile
from pathlib import Path
from typing import Union

from fsspec import AbstractFileSystem
from safetensors.flax import save, save_file

from safejax.typing import ParamsDictLike, PathLike
from safejax.utils import flatten_dict


def serialize(
    params: ParamsDictLike,
    include_objax_variables: bool = False,
    filename: Union[PathLike, None] = None,
    fs: Union[AbstractFileSystem, None] = None,
) -> Union[bytes, PathLike]:
    """
    Serialize JAX, Flax, Haiku, or Objax model params from either `FrozenDict`, `Dict`, or `VarCollection`.

    If `filename` is not provided, the serialized model is returned as a `bytes` object,
    otherwise the model is saved to the provided `filename` and the `filename` is returned.

    Args:
        params: A `FrozenDict`, a `Dict` or a `VarCollection` containing the model params.
        include_objax_variables: Whether to include `objax.Variable` objects in the serialized model params.
        filename: The path to the file where the model params will be saved.
        fs: The filesystem to use to save the model params. Defaults to `None`.

    Returns:
        The serialized model params as a `bytes` object or the path to the file where the model params were saved.
    """
    params = flatten_dict(
        params=params, include_objax_variables=include_objax_variables
    )

    if filename:
        if not isinstance(filename, (str, Path)):
            raise ValueError(
                "If `filename` is provided (not `None`), it must be a `str` or a"
                f" `pathlib.Path` object, not {type(filename)}."
            )
        if fs and fs.protocol != "file":
            if not isinstance(fs, AbstractFileSystem):
                raise ValueError(
                    "`fs` must be a `fsspec.AbstractFileSystem` object or `None`,"
                    f" not {type(fs)}."
                )
            temp_filename = tempfile.NamedTemporaryFile(
                mode="wb", suffix=".safetensors", delete=False
            )
            try:
                temp_filename.write(save(tensors=params))
            finally:
                temp_filename.close()
                fs.put_file(lpath=temp_filename.name, rpath=filename)
                os.remove(temp_filename.name)
        else:
            if fs and fs.protocol == "file":
                filename = Path(fs._strip_protocol(filename))
            else:
                filename = filename if isinstance(filename, Path) else Path(filename)
            if not filename.exists or not filename.is_file:
                raise ValueError(
                    f"`filename` must be a valid file path, not {filename}."
                )
            save_file(tensors=params, filename=filename.as_posix())
        return filename

    return save(tensors=params)



import warnings
from typing import Any, Dict, Union

import numpy as np
from flax.core.frozen_dict import FrozenDict
from jax import numpy as jnp
from objax.variable import BaseState, BaseVar, RandomState, StateVar, TrainRef, TrainVar

from safejax.typing import JaxDeviceArrayDict, ObjaxDict, ParamsDictLike

OBJAX_VARIABLES = {
    "BaseVar": BaseVar,
    "BaseState": BaseState,
    "RandomState": RandomState,
    "TrainRef": TrainRef,
    "StateVar": StateVar,
    "TrainVar": TrainVar,
}
OBJAX_VARIABLE_SEPARATOR = "::"


def flatten_dict(
    params: ParamsDictLike,
    key_prefix: Union[str, None] = None,
    include_objax_variables: bool = False,
) -> Union[Dict[str, np.ndarray], Dict[str, jnp.DeviceArray]]:
    """
    Flatten a `Dict`, `FrozenDict`, or `VarCollection`, for more detailed information on
    the supported input types check `safejax.typing.ParamsDictLike`.

    Note:
        This function is recursive to explore all the nested dictionaries,
        and the keys are being flattened using the `.` character. So that the
        later de-nesting can be done using the `.` character as a separator.

    Reference at https://gist.github.com/Narsil/d5b0d747e5c8c299eb6d82709e480e3d

    Args:
        params: A `Dict`, `FrozenDict`, or `VarCollection` with the params to flatten.
        key_prefix: A prefix to prepend to the keys of the flattened dictionary.
        include_objax_variables:
            A boolean indicating whether to include the `objax.variable` types in
            the keys of the flattened dictionary.

    Returns:
        A `Dict` containing the flattened params as level-1 key-value pairs.
    """
    flattened_params = {}
    for key, value in params.items():
        key = f"{key_prefix}.{key}" if key_prefix else key
        if isinstance(value, (BaseVar, BaseState)):
            if include_objax_variables:
                key = f"{key}{OBJAX_VARIABLE_SEPARATOR}{type(value).__name__}"
            value = value.value
        if isinstance(value, (jnp.DeviceArray, np.ndarray)):
            flattened_params[key] = value
            continue
        if isinstance(value, (Dict, FrozenDict)):
            flattened_params.update(
                flatten_dict(
                    params=value,
                    key_prefix=key,
                    include_objax_variables=include_objax_variables,
                )
            )
    return flattened_params


# Note that this function has a less restrictive type hint than the `flatten_dict` function.
# This is because the `unflatten_dict` function is generic, and it can be used to unflatten
# any `Dict` where the keys are expanded using the `.` character as a separator.
def unflatten_dict(params: Dict[str, Any]) -> Dict[str, Any]:
    """
    Unflatten a `Dict` where the keys should be expanded using the `.` character
    as a separator.

    Note:
        If the params where serialized from a `VarCollection` object, then the
        `objax.variable` types are included in the keys, and since this function
        just unflattens the dictionary without `objax.variable` casting, those
        variables will be ignored and unflattened normally. Anyway, when deserializing
        `objax` models you should use `safejax.objax.deserialize` or just use the
        function params in `safejax.deserialize`: `requires_unflattening=False` and
        `to_var_collection=True`.

    Reference at https://stackoverflow.com/a/63545677.

    Args:
        params: A `Dict` containing the params to unflatten by expanding the keys.

    Returns:
        An unflattened `Dict` where the keys are expanded using the `.` character.
    """
    unflattened_params = {}
    warned_user = False
    for key, value in params.items():
        unflattened_params_tmp = unflattened_params
        if not warned_user and OBJAX_VARIABLE_SEPARATOR in key:
            warnings.warn(
                "The params were serialized from a `VarCollection` object, "
                "so the `objax.variable` types are included in the keys, "
                "and since this function just unflattens the dictionary "
                "without `objax.variable` casting, those variables will be "
                "ignored and unflattened normally. Anyway, when deserializing "
                "`objax` models you should use `safejax.objax.deserialize` "
                "or just use the function params in `safejax.deserialize`: "
                "`requires_unflattening=False` and `to_var_collection=True`."
            )
            warned_user = True
        key = (
            key.split(OBJAX_VARIABLE_SEPARATOR)[0]
            if OBJAX_VARIABLE_SEPARATOR in key
            else key
        )
        subkeys = key.split(".")
        for subkey in subkeys[:-1]:
            unflattened_params_tmp = unflattened_params_tmp.setdefault(subkey, {})
        unflattened_params_tmp[subkeys[-1]] = value
    return unflattened_params


def cast_objax_variables(
    params: Dict[str, jnp.DeviceArray]
) -> Union[JaxDeviceArrayDict, ObjaxDict]:
    """
    Cast the `jnp.DeviceArray` to their corresponding `objax.variable` types.

    Note:
        This function may return the same `params` if no `objax.variable` types
        are found in the keys.

    Args:
        params: A `Dict` containing the params to cast.

    Raises:
        ValueError: If the params were not serialized from a `VarCollection` object.

    Returns:
        A `Dict` containing the keys without the variable name, and the values
        with the `objax.variable` objects with `.value` assigned from the
        `jnp.DeviceArray`.
    """
    casted_params = {}
    for key, value in params.items():
        if OBJAX_VARIABLE_SEPARATOR not in key:
            raise ValueError(
                "The params were not serialized from a `VarCollection` object, since"
                " the type has not been included as part of the key using"
                f" `{OBJAX_VARIABLE_SEPARATOR}` as separator at the end of the key."
                " Returning the same params without casting the `jnp.DeviceArray` to"
                " `objax.variable` types."
            )
        key, objax_var_type = key.split(OBJAX_VARIABLE_SEPARATOR)
        casted_params[key] = OBJAX_VARIABLES[objax_var_type](value)
    return casted_params



from time import perf_counter

import jax
from flax.serialization import to_bytes
from flaxmodels.resnet import ResNet50
from jax import numpy as jnp

from safejax import serialize

resnet50 = ResNet50()
params = resnet50.init(jax.random.PRNGKey(42), jnp.ones((1, 224, 224, 3)))


start_time = perf_counter()
for _ in range(100):
    serialize(params)
end_time = perf_counter()
print(f"safejax (100 runs): {end_time - start_time:0.4f} s")

start_time = perf_counter()
for _ in range(100):
    to_bytes(params)
end_time = perf_counter()
print(f"flax (100 runs): {end_time - start_time:0.4f} s")



from functools import partial

from safejax.core.load import deserialize
from safejax.core.save import serialize  # noqa: F401

# `flax` expects either a `Dict[str, Any` or a `FrozenDict`, but for robustness we are
# setting `freeze_dict` to `True` to restore a `FrozenDict` which contains the params
# frozen to avoid any accidental mutation.
deserialize = partial(deserialize, freeze_dict=True)



import sys

import jax
import jax.numpy as jnp
from flax import linen as nn
from flax.serialization import to_bytes

from safejax import serialize


class SingleLayerModel(nn.Module):
    features: int

    @nn.compact
    def __call__(self, x):
        x = nn.Dense(features=self.features)(x)
        return x


model = SingleLayerModel(features=1)

rng = jax.random.PRNGKey(0)
params = model.init(rng, jnp.ones((1, 1)))


def serialization_safejax():
    _ = serialize(params)


def serialization_flax():
    _ = to_bytes(params)


if __name__ == "__main__":
    if len(sys.argv) < 2:
        raise ValueError("Please provide a function name to run as an argument")
    if sys.argv[1] not in globals():
        raise ValueError(f"Function {sys.argv[1]} not found")
    globals()[sys.argv[1]]()



from safejax.core.load import deserialize  # noqa: F401
from safejax.core.save import serialize  # noqa: F401

# Nothing here as `dm-haiku` works with the default behavior of both
# `safejax.core.load.deserialize` and `safejax.core.save.serialize`. But
# placing this here for consistency with the other frameworks.
